
AC_DEFUN(AC_CHECK_SIAG,
[
dnl We want to use gmatch where fnmatch is not available
AC_CHECK_FUNC(fnmatch, AC_DEFINE(HAVE_FNMATCH),
	AC_CHECK_LIB(gen, gmatch)
	AC_CHECK_FUNC(gmatch, AC_DEFINE(HAVE_GMATCH)))

dnl Somehow we must be able to set NDBM even if no libdb exists (Solaris)
dnl Try named lib first, then libdb, then libgdbm, then libndbm
define(siag_CHECK_NDBM,
[AC_CHECK_LIB($MAYBE_LIBNDBM, dbm_open, LIBDB=-l$MAYBE_LIBNDBM,
	AC_CHECK_LIB(db, dbm_open, LIBDB="-ldb",
		AC_CHECK_LIB(gdbm, dbm_open, LIBDB="-lgdbm",
			AC_CHECK_LIB(ndbm, dbm_open, LIBDB="-lndbm"))))
if test "$LIBDB" != ""; then
	echo "Found ndbm in $LIBDB"
	LIBS="$LIBDB $LIBS"
fi
AC_CHECK_FUNC(dbm_dirfno, AC_DEFINE(HAVE_DBM_DIRFNO))
AC_CHECK_FUNC(dbm_pagfno, AC_DEFINE(HAVE_DBM_PAGFNO))
AC_CHECK_FUNC(dbm_rdonly, AC_DEFINE(HAVE_DBM_RDONLY))
AC_CHECK_FUNC(dbm_open, NDBM="-DNDBM")
])

AC_ARG_WITH(ndbm,
	[  --with-ndbm[=libname]   enable (named) ndbm library [default: gdbm]],
	[ case $withval in
	yes)
		MAYBE_LIBNDBM="gdbm"
		;;
	no)
		;;
	*)
		MAYBE_LIBNDBM=$withval
		;;
	esac
	
	if test "$withval" != "no"; then
		siag_CHECK_NDBM
	fi ],
dnl	[ MAYBE_LIBNDBM="gdbm"
dnl	siag_CHECK_NDBM ])
)

AC_SUBST(NDBM)

AC_CHECK_LIB(m, main)

dnl Check what math functions we have
AC_CHECK_FUNC(acosh, AC_DEFINE(HAVE_ACOSH))
AC_CHECK_FUNC(asinh, AC_DEFINE(HAVE_ASINH))
AC_CHECK_FUNC(log1p, AC_DEFINE(HAVE_LOG1P))
AC_CHECK_FUNC(expm1, AC_DEFINE(HAVE_EXPM1))
AC_CHECK_FUNC(cbrt, AC_DEFINE(HAVE_CBRT))
AC_CHECK_FUNC(drem, AC_DEFINE(HAVE_DREM))

AC_CHECK_LIB(crypt, main, [LIBSYSDEP="-lcrypt"])
if test "$LIBSYSDEP" = ""; then
	AC_CHECK_LIB(posix4, main, [LIBSYSDEP="-lposix4"])
	if test "$LIBSYSDEP" = ""; then
		LIBSYSDEP=""
	fi
fi
AC_SUBST(LIBSYSDEP)

dnl These are necessary for SIOD on Solaris and probably Irix.
AC_CHECK_LIB(nsl, main)
AC_CHECK_LIB(socket, connect)

dnl SIOD
AC_CHECK_HEADERS(sys/mode.h fnmatch.h crypt.h sys/mkdev.h fcntl.h)

AC_CHECK_FUNCS(setpwfile getrlimit getrusage fnmatch \
		usleep lchown strptime putpwent getwd)

dnl Guile and Tcl

dnl AC_CHECK_LIB(dl, main, LIBDL="-ldl")
dnl AC_SUBST(LIBDL)

dnl Empty libdl in case we are making static binaries
dnl LIBDL=""

define(siag_CHECK_TCL,
[AC_CHECK_LIB(dl, main, LIBDL="-ldl")
AC_CHECK_LIB($MAYBE_LIBTCL, Tcl_Main, LIBTCL=-l$MAYBE_LIBTCL, , $LIBMATH $LIBDL)
if test "$LIBTCL" != ""; then
	TCL="-DTCL"
fi])

AC_ARG_WITH(tcl,
	[  --with-tcl[=libname]    enable (named) tcl library [default: tcl]],
	[ case $withval in 
	yes)
		MAYBE_LIBTCL="tcl"
		;;
	no)
		;;
	*)
		MAYBE_LIBTCL=$withval
		;;
	esac

	 if test "$withval" != "no"; then
		siag_CHECK_TCL
	fi ],
dnl	[ MAYBE_LIBTCL="tcl"
dnl	  siag_CHECK_TCL ])
)
AC_SUBST(TCL)
AC_SUBST(LIBTCL)

AC_ARG_WITH(guile,
	[  --with-guile            enable guile commands and expressions],
	[ if test "$withval" != "no"; then
		AC_CHECK_LIB(guile, gh_enter, [LIBGUILE=`guile-config link`], , `guile-config link`)
		if test "$LIBGUILE" != ""; then
			GUILE="-DGUILE"
		fi
	fi ],
dnl	[ AC_CHECK_LIB(guile, gh_enter, [LIBGUILE=`guile-config link`], , `guile-config link`)
dnl	if test "$LIBGUILE" != ""; then
dnl		GUILE="-DGUILE"
dnl	fi ])
)
AC_SUBST(GUILE)
AC_SUBST(LIBGUILE)
AC_SUBST(LIBDL)

AC_ARG_WITH(python,
	[  --with-python           enable python commands and expressions],
	[ if test "$withval" != "no"; then
		AC_CHECK_LIB(pthread, main)
		AC_CHECK_LIB(python1.5, PyFloat_AsDouble, [LIBPYTHON="-lpython1.5"], ,)
		if test "$LIBPYTHON" != ""; then
			PYTHON="-DPYTHON"
		fi
	fi ])
AC_SUBST(PYTHON)
AC_SUBST(LIBPYTHON)

dnl No Perl
LIBPERL=""
AC_SUBST(LIBPERL)
AC_SUBST(PERL)
AC_SUBST(PERLINC)

dnl This is insane...
ac_save_LDFLAGS="$LDFLAGS"
LDFLAGS="$LDFLAGS $X_LIBS"
AC_CHECK_LIB(Xext, main, [LIBXEXT="-lXext"], , $X_PRE_LIBS $X_EXTRA_LIBS $X_EXTRA_LIBS -lXt -lX11)
AC_SUBST(LIBXEXT)

AC_CHECK_LIB(Xmu, main, [LIBXMU="-lXmu"], , $LIBXEXT $X_PRE_LIBS $X_EXTRA_LIBS $X_EXTRA_LIBS -lXt -lX11)
AC_SUBST(LIBXMU)
AC_CHECK_LIB(X11, XCreateIC, AC_DEFINE(HAVE_XCREATEIC), , $x_libs)
LDFLAGS=$ac_save_LDFLAGS

if test "$GCC" = "yes"; then
	CFLAGS="-Wall $CFLAGS"
fi

if test "$XAWLIB" = ""; then
	XAWLIB="-lXaw3d"
fi
AC_SUBST(XAWLIB)

dnl What curses lib is available? Try in this order:
dnl 0. use termcap if available
dnl 1. ncurses
dnl 2. curses
ac_save_LDFLAGS="$LDFLAGS"
AC_CHECK_LIB(termcap, main, LIBTERMCAP="-ltermcap")
AC_CHECK_LIB(ncurses, wgetch, LIBCURSES="-lncurses $LIBTERMCAP",
	AC_CHECK_LIB(curses, wgetch, LIBCURSES="-lcurses $LIBTERMCAP", , $LIBTERMCAP),
	$LIBTERMCAP)
LDFLAGS="$LDFLAGS $LIBCURSES"
AC_CHECK_FUNCS(resizeterm beep immedok keypad)
AC_SUBST(LIBCURSES)
LDFLAGS=$ac_save_LDFLAGS

dnl Look for T1lib including its X part
AC_CHECK_LIB(t1, main, LIBT1="-lt1")
if test "$LIBT1" != ""; then
	AC_DEFINE(HAVE_LIB_T1)
fi
AC_SUBST(LIBT1)
AC_CHECK_LIB(t1x, main, LIBT1X="-lt1x", , $LIBT1 $X_LIBS -lX11)
if test "$LIBT1X" != ""; then
	AC_DEFINE(HAVE_LIB_T1X)
fi
AC_SUBST(LIBT1X)
])
